#include <iostream>

#include <arc/xrsl.h>

using namespace std; 

int main( ) {
    
	// Constructor + str
	Xrsl xrsl("&(executable=/bin/ls)(arguments=-la)");
	cout << "The xrsl:" << endl << xrsl.str() << endl; 

	// Copy constructor
	Xrsl other(xrsl); 
	cout << "A copy of the above xrsl:" << endl << other.str() << endl; 

	// AddSimpleRelation
	xrsl.AddSimpleRelation("middleware", operator_gteq, "nordugrid-0.5");
	xrsl.AddSimpleRelation("middleware", operator_gteq, "globus-2.4"); 

	xrsl.AddSimpleRelation("count", operator_eq, "1"); 
	xrsl.AddSimpleRelation("disk", operator_gteq, "500"); 
	cout << "AddSimpleRelation:" << endl << xrsl.str() << endl; 

	// GetSingleValue
	cout << "GetSingleValue:" << endl; 
	string executable = xrsl.GetRelation("executable").GetSingleValue();
	cout <<  "executable = \"" << executable << "\"" << endl; 

	// GetAll
	list<XrslRelation> rellist = xrsl.GetAllRelations("middleware");
	cout << "Found relations:" << endl;
	list<XrslRelation>::iterator it;
	for (it = rellist.begin(); it != rellist.end(); it++) {
		cout << it->str() << endl;
	}
	// RemoveSimpleRelation
	xrsl.RemoveRelation("count"); 
	cout << "RemoveRelation:" << endl << xrsl.str() << endl; 

	// Error handling: 
	// Remove relation that does not exist: 
	cout << "Removing non-existing relation:" << endl; 
	try {
		xrsl.RemoveRelation("does_not_exist"); 
	} catch (XrslError e) {
		cerr << "Caught exception: " <<  e.what() << endl; 
	}
    
	// Get non-existing relation
	cout << "Getting non-existing relation (nothing):" << endl; 
	try {
		string nothing = xrsl.GetRelation("nothing").GetSingleValue();
	} catch (XrslError e) {
		cerr << "Caught exception: " <<  e.what() << endl; 
	}

	// Add duplicate of relation
	cout << "Adding duplicate of relation:" << endl; 
	try {
		xrsl.AddSimpleRelation("executable", operator_eq, "/bin/true"); 
	} catch (XrslError e) {
		cerr << "Caught exception: " <<  e.what() << endl; 
	}

	cout << "The final xrsl:" << endl << xrsl.str();

	// Test list relation handling
	Xrsl x2("&(executable=program)"); 

	// AddListRelation
	list<string> data; 
	data.push_back("data"); 
	data.push_back("www.someserver.com"); 
	list<string> parameters; 
	parameters.push_back("parameters"); 
	parameters.push_back("file:///tmp/my_parameters"); 

	list<list<string> > input;
	input.push_back(data); 
	input.push_back(parameters); 

	cout << endl <<  "Testing list manipulation, adding inputfiles" << endl; 
	XrslRelation listrel("inputfiles", operator_eq, input);
	x2.AddRelation(listrel);

	std::cout << x2.str() << std::endl;
	// GetListRelation
	cout << "Getting inputfiles:" << endl; 
	list<list<string> > simple; 
	simple = x2.GetRelation("inputfiles").GetDoubleListValue();
	
	for(list<list<string> >::iterator outer = 
	    simple.begin(); outer != simple.end(); outer++) {
		for(list<string>::iterator inner = outer->begin(); 
		    inner != outer->end(); inner++) {
			cout << *inner << "  "; 
		}
		cout << endl; 
	}

	cout << "Xrsl is now: " << endl << x2.str() << endl; 
	// Try to add relation that already exits
	cout << "Trying to add (non-forcing) duplicate of relation" << endl; 
	try {
		x2.AddRelation(listrel, false);
	} catch (XrslError e) {
		cerr << e.what() << endl; 
	}
	cout << "Trying to get non-existing relation" << endl; 
	// Try to get relation that does not exist
	try {
		x2.GetRelation("no_such_list"); 
	} catch (XrslError e) {
		cerr << e.what() << endl; 
	}
    
	// RemoveRelation - works on lists as 
	cout << "Removing relations (non-existing and existing)" << endl;
	try {
		x2.RemoveRelation("inputfiled"); 
	} catch (XrslError e) {
		cerr << e.what() << endl; 
	}
	try {
		x2.RemoveRelation("inputfiles"); 
		cout << "Removed relation \"inputfiles\"" << endl; 
	} catch (XrslError e) {
		cerr << e.what() << endl; 
	}
	cout << "Final x2: " << endl << x2.str() << endl; 

	// Ensure that no-argument constructor works
	cout << endl << "Creating empty Xrsl x3" << endl;  
	Xrsl x3; 
	cout << "x3: " << endl << x3.str() << endl; 
	cout << "Inserting one attribute" << endl;
	x3.AddSimpleRelation("something", operator_eq, "some_value"); 
	cout << "x3: " << endl << x3.str() << endl; 

}

