/*
 * Created on Aug 1, 2005
 *
 *
 */
package se.umu.jobsubmission.impl.broker;

import java.util.HashMap; 
import java.util.Map; 

import org.ggf.schemas.jsdl._2005._11.jsdl.JobDefinition_Type;

import org.globus.gsi.GlobusCredential;
import org.globus.mds.glue.GLUECERPType;

import se.umu.jobsubmission.JobPreferencesType;
import se.umu.jobsubmission.util.Worker;

import org.apache.log4j.Logger;

/**
 * @author tordsson
 *
 * Determines stage in time and job wait time for a job. 
 * instances of this class are executed in a thread pool 
 * when reordering the resources according estimate job start time. 
 */
public class EarliestCompletionWorker extends Worker {

	private static Logger logger = Logger.getLogger(
			EarliestCompletionWorker.class); 
	
	private Predictor stageInTimePredictor; 
	private Predictor startTimePredictor; 
	private Predictor runTimePredictor; 
	private Predictor stageOutTimePredictor; 
	
	private JobDefinition_Type jobDescription; 
	private JobPreferencesType preferences; 
	private Map fileSizes; 	
	private GlobusCredential credential;
	
	/**
	 * @param stageInTimePredictor
	 * @param startTimePredictor
	 * @param runTimePredictor
	 * @param stageOutTimePredictor
	 * @param jobDescription
	 * @param preferences
	 * @param fileSizes
	 */
	public EarliestCompletionWorker(Object resource,
			JobDefinition_Type jobDescription,
			JobPreferencesType preferences, 
			Map fileSizes, 
			Predictor stageInTimePredictor,
			Predictor startTimePredictor, 
			Predictor runTimePredictor,
			Predictor stageOutTimePredictor, 
			GlobusCredential credential) {

		super(resource); 
		this.stageInTimePredictor = stageInTimePredictor;
		this.startTimePredictor = startTimePredictor;
		this.runTimePredictor = runTimePredictor;
		this.stageOutTimePredictor = stageOutTimePredictor;
		this.jobDescription = jobDescription;
		this.preferences = preferences;
		this.fileSizes = fileSizes;
		this.credential = credential; 
		
	}



    public void doWork() {
    	
    	int startTime = 0; 
		int runTime = 0; 
		int stageInTime = 0; 
		int stageOutTime = 0; 
		String resourceName = null; 
		
		resourceName= ((GLUECERPType)this.workUnit).getCluster(0).
			getSubCluster(0).getHost(0).getName();     		
		logger.debug("Determining TTD for " + resourceName); 
		
		// stageInTime, 
		stageInTime = this.stageInTimePredictor.predict(jobDescription, 
				this.workUnit, 
				preferences, 
				fileSizes, 
				this.credential); 
    	
		// runTime: 
		runTime = this.runTimePredictor.predict(jobDescription, 
				this.workUnit, 
				preferences, 
				null, 
				this.credential);   
		
		// startTime: 
		// pass runTime and stageInTime to startTimePredictor
		Map context = new HashMap(); 
		context.put("runTime", new Integer(runTime)); 
		context.put("stageInTime", new Integer(stageInTime)); 			
		startTime = this.startTimePredictor.predict(jobDescription, 
				this.workUnit, 
				preferences, 
				context, 
				this.credential);     			
		
		// stageOutTime
		stageOutTime = this.stageOutTimePredictor.predict(jobDescription, 
				this.workUnit, 
				preferences, 
				null, 
				this.credential); 						
		
		int ttd = startTime + runTime + (stageOutTime * 60); 
		logger.debug("TTD for resource " + resourceName + " is "
				+ startTime + " + " + runTime + " + " + (stageOutTime * 60) + " = "
				+ ttd); 
		
		// Store time with resource in result set. 
		RankedResource resource = new RankedResource(
				(GLUECERPType)this.workUnit, 
				ttd); 
		this.resultSet.add(resource); 
    	
    }
	
}
