/*
 * Created on Oct 19, 2005
 *
 * TODO Implement it, then write some documentation
 *
 */
package se.umu.jobsubmission.impl.dispatcher;

import java.io.File;
import java.net.URI;
import java.util.Iterator;
import java.util.List;
import java.util.Vector;

import org.apache.axis.MessageContext; 
import org.apache.log4j.Logger;
import org.globus.mds.glue.GLUECERPType;
import org.globus.rsl.AbstractRslNode;
import org.globus.rsl.NameOpValue;
import org.globus.rsl.Value;

import org.globus.gram.GramJob;
import org.globus.gsi.GlobusCredential;
import org.globus.gsi.gssapi.GlobusGSSCredentialImpl;
import org.ietf.jgss.GSSCredential;

import se.umu.jobsubmission.util.SubmitterUtils; 


import se.umu.jobsubmission.util.JobDescriptionTranslator;
/**
 * @author tordsson
 *
 * TODO Add a proper description of this class. 
 */
public class GT2Dispatcher implements Dispatcher {

    private static Logger logger = Logger.getLogger(GT2Dispatcher.class); 
    private JobDescriptionTranslator descriptionTranslator; 

    public GT2Dispatcher() {
        descriptionTranslator = (JobDescriptionTranslator) new JSDL2GT2RSL(); 
    }

    /** 
     * Constructs a new value with the replaced fileURL to 
     * the locally stored input files. 
     * @param filePath
     * @param resourceName
     * @param clientURL
     * @return
     */
    private List buildValue(String filePath, 
            String resourceName, 
            String clientURL) {

        List value = new Vector(); 
        File f = new File(filePath); 
        String URL = "gsiftp://" 
            + clientURL
            + ":2811" 
            + f.getAbsolutePath(); 
        value.add(new Value(URL));
        value.add(new Value(resourceName)); 

        return value; 
    }


    /**
     * Modifies the rsl to ensure that the GM also uploads locally stored
     * files... 
     * @param rsl
     * @param clientURL
     * @throws Exception
     */
    public void fixInputFiles(AbstractRslNode rsl, String clientURL) 
        throws Exception {

            NameOpValue inputfiles = rsl.getParam("file_stage_in"); 

            // Values are stored as list (pair) of lists 
            if(inputfiles != null) {

                List values = inputfiles.getValuesAsStrings(true); 
                Iterator iter = values.iterator(); 

                List addedFiles = new Vector(); 
                List removedFiles = new Vector(); 

                // Extract files stored on client host, change the URL of them 
                // to gsiftp://clienturl/... causing the JM to upload files. 
                // This is done because the JSS and the client need not share a
                // file system, making it impossible for the JSS to upload 
                // locally stored input files to the resource. 

                while(iter.hasNext()) {
                    List l; 
                    try {
                        l = (List) iter.next(); 
                    }
                    catch (ClassCastException e) {
                        throw new Exception("inputfiles must be a list of lists!"); 
                    }

                    // Format is (remote local)  
                    // N.B: localURL is local on the resource, 
                    //      remoteURL is local to the client host. 
                    String remote = (String) l.get(0); 
                    String local = (String) l.get(1); 

                    URI remoteURI = new URI(remote); 
                    boolean removeLater = false; 

                    // Scheme exist, but is file:// 
                    if("file".equals(remoteURI.getScheme())) {
                        logger.debug("Protocol is [local]file://"); 
                        removeLater = true; 
                        List value = buildValue(remoteURI.getPath(), 
                                local, 
                                clientURL); 
                        addedFiles.add(value); 
                    }

                    if(removeLater) {
                        List value = new Vector(); 
                        value.add(new Value(remote)); 
                        value.add(new Value(local)); 
                        removedFiles.add(value); 
                    }
                }

                // Now, remove all files that were scheduled for transfer 
                // by the dispatcher. 
                for(int i = 0; i < addedFiles.size(); i++) {
                    inputfiles.add((List)addedFiles.get(i));
                }
                for(int i = 0; i < removedFiles.size(); i++) {
                    inputfiles.remove(removedFiles.get(i)); 
                }
                if(inputfiles.getValues().size() == 0) {
                    NameOpValue rel = rsl.removeParam("file_stage_in");  
                    if(rel != null)
                        logger.debug("removed " + rel.getAttribute()); 
                }
            }
        }

    public Object requestJob(Object resource,   
            Object jobDescription, 
            Object reservation, 
            String clientURL, 
			boolean clientStagesInput, 
			GlobusCredential credential) throws Exception {

        String resourceURI = null; 
        try {

            logger.info("Dispatching job using: " + GT2Dispatcher.class); 

            // Get GRAM gatekeeper URI
            GLUECERPType resourceInfo = (GLUECERPType) resource; 
            resourceURI = resourceInfo.getCluster(0).getUniqueID(); 

            // Cheat a bit in testing: 
            // resourceURI = "i101.hpc2n.umu.se:2119"; 

            // Translate job description
            AbstractRslNode rsl = (AbstractRslNode) descriptionTranslator.translate(
                    jobDescription); 

            if(!clientStagesInput) {
            	logger.info("Preparing staging of local input files"); 
            	fixInputFiles(rsl, clientURL); 
            }
            else {
            	logger.info("No staging performed (it is handled by client)"); 
            }

            // Advertised by LCG MDS: 
            // i101.hpc2n.umu.se:2119/jobmanager-lcgpbs-ngrid
            // But real job manager name is:
            // i101.hpc2n.umu.se:2119/jobmanager-lcgpbs

            logger.debug("Resource job manager is: : " + resourceURI);	    	

            // N.B: This ensures compatability with the LCG
            // job manager, which, for some reason, is 
            // incorrectly named in the LCG MDS. 
            int index = resourceURI.indexOf("-lcgpbs-");
            if(index > 0) {
                logger.debug("This is a LCG resource, fixing URI"); 
                resourceURI = resourceURI.substring(0, index + "-lcgpbs".length());
                logger.debug("ResourceURI is now: " + resourceURI); 
            }	    	

            logger.debug("Submitted RSL:\n" + rsl.toString()); 

            // Create job and set rsl. 
            GramJob job = new GramJob(rsl.toRSL(true)); 

            // Delegate proxy to job: 
            GSSCredential cred = new GlobusGSSCredentialImpl(credential, 
                    GSSCredential.INITIATE_AND_ACCEPT);             
            job.setCredentials(cred); 

            // Submit job: 
            job.request(resourceURI); 

            // Get job id: 
            String jobID = job.getIDAsString(); 

            logger.debug("Job submitted with id: " + jobID); 

            return jobID; 
        } catch (Exception e) {
            logger.error("Error submitting job to resource: " + resourceURI); 
            throw e; 
        }
    }
}
