package org.nordugrid.job.gridftp;

import java.io.InputStream;
import java.io.OutputStream;
import java.util.Vector;
import org.globus.ftp.GridFTPClient;
import org.globus.util.GlobusURL;
import org.ietf.jgss.GSSManager;
import org.ietf.jgss.GSSCredential;
import org.apache.log4j.Logger;
import org.globus.gsi.gssapi.GlobusGSSManagerImpl;
import org.globus.gsi.GlobusCredential;
import org.globus.gsi.gssapi.GlobusGSSCredentialImpl;
import org.globus.ftp.DataSourceStream;
import org.globus.ftp.DataSinkStream;
import org.globus.ftp.Marker;
import org.globus.ftp.MarkerListener;
import org.globus.ftp.MlsxEntry;
import org.globus.ftp.exception.ClientException;
import org.globus.ftp.exception.ServerException;

class DummyMarkerListener implements MarkerListener {

    public DummyMarkerListener() {
    }

    public void markerArrived(Marker m) {
    }

}

public class ARCGridFTPJobInterface {
	
	private static Logger logger = 
		Logger.getLogger(ARCGridFTPJobInterface.class); 

    GlobusURL url = null; 
	
    protected GridFTPClient client = null;

    public ARCGridFTPJobInterface(String u) throws ARCGridFTPJobException {
       try {
           url = new GlobusURL(u);
       } catch(java.net.MalformedURLException err) {
           throw new ARCGridFTPJobException("Bad URL "+u+": "+err.getMessage());
       }
       if(!url.getProtocol().equals("gsiftp")) {
           String err = new String("Protocol "+url.getProtocol()+
           		" not supported for job submission");
           url=null;
           throw new ARCGridFTPJobException(err);
       }
    }

    protected void finalize() {
        Disconnect();
    }

    public void Connect() throws ARCGridFTPJobException {
    	Connect(null, null);
    }
    
    public void Connect(String proxylocation) throws ARCGridFTPJobException {
    	logger.debug("Proxy: " + proxylocation); 
    	Connect(proxylocation,null);
    }
    
    public void Connect(GlobusCredential gc) throws ARCGridFTPJobException {
    	Connect(null, gc);
    }
    public void Connect(String proxylocation, GlobusCredential gc) throws ARCGridFTPJobException {    	
        if(client == null) {
            if(url == null) {
                throw new ARCGridFTPJobException("URL not defined");
            }
            try {
                int port = url.getPort();
                if(port == -1) {
                  port = GlobusURL.getPort(url.getProtocol());
                }
                client = new GridFTPClient(url.getHost(),port);
                GSSManager manager = GlobusGSSManagerImpl.getInstance();
                GlobusCredential gcred = gc;
                if (proxylocation != null) {
                	try {	
                		gcred = new GlobusCredential(proxylocation);
                	} catch (Exception e) {
                		logger.error("reading proxy from file failed, continuing");
                	}
                }
                GSSCredential cred = null;
                if (gcred != null) {
                	GlobusGSSCredentialImpl globusgsscred = new GlobusGSSCredentialImpl(gcred,0);
                	cred =  globusgsscred;
                } else {
                      cred =  manager.createCredential(GSSCredential.INITIATE_ONLY);
                }              
                client.authenticate(cred);
                client.changeDir("/"+url.getPath());
            } catch(java.io.IOException err) {
                client=null;
                throw new ARCGridFTPJobException("IO error: "+err.getMessage());
            } catch(ServerException err) {
                client=null;
                throw new ARCGridFTPJobException("Communication with server failed: "+err.getMessage());
            } catch(org.ietf.jgss.GSSException err) {
                client=null;
                throw new ARCGridFTPJobException("Failed to use credentials: "+err.getMessage());
            }
        }
    }

    public void Disconnect() {
        if(client != null) {
            try {
                client.close();
            } catch(ServerException err) {
                client = null;
            } catch(java.io.IOException err) {
                client = null;
            }
            client = null;
        }
    }

	//submit using default proxy
	public String Submit() throws ARCGridFTPJobException {
		return Submit(null, null);
	}

	
	//submit using a proxy location e.g. /tmp/proxy
	public String Submit(String proxylocation) throws ARCGridFTPJobException {
		return Submit(proxylocation, null);
	}
	
	//submit using a credential object
	public String Submit(GlobusCredential proxy) throws ARCGridFTPJobException {
			return Submit(null, proxy);
		}
		
		
    private String Submit(String proxyloc, GlobusCredential proxy) throws ARCGridFTPJobException {
        if(client == null) {
        	if (proxyloc != null) Connect(proxyloc);
        	else if (proxy != null) Connect(proxy);
        	else Connect();
        } 
        String id = new String("");
        try {
            client.changeDir("new");
            id = client.getCurrentDir();
            int p = id.lastIndexOf('/');
            if(p != -1) {
                id = id.substring(p+1);
            };
        } catch(java.io.IOException err) {
            Disconnect();
            throw new ARCGridFTPJobException("IO error: "+err.getMessage());
        } catch(org.globus.ftp.exception.ServerException err) {
            Disconnect();
            throw new ARCGridFTPJobException("Communication with server failed: "+err.getMessage());
        }
        return id;
    }

    public void Cancel(String id) throws ARCGridFTPJobException {
        if(client == null) Connect();
        try {
            client.deleteFile("/"+url.getPath()+"/"+id);
        } catch(java.io.IOException err) {
            Disconnect();
            throw new ARCGridFTPJobException("IO error: "+err.getMessage());
        } catch(org.globus.ftp.exception.ServerException err) {
            Disconnect();
            throw new ARCGridFTPJobException("Communication with server failed: "+err.getMessage());
        }
    }

    public void Clean(String id) throws ARCGridFTPJobException {
        if(client == null) Connect();
        try {
            client.deleteDir("/"+url.getPath()+"/"+id);
        } catch(java.io.IOException err) {
            Disconnect();
            throw new ARCGridFTPJobException("IO error: "+err.getMessage());
        } catch(org.globus.ftp.exception.ServerException err) {
            Disconnect();
            throw new ARCGridFTPJobException("Communication with server failed: "+err.getMessage());
        }
    }

    public void Upload(InputStream in,String name) throws ARCGridFTPJobException {
        if(client == null) Connect();
        try{
            client.setPassiveMode(true);
            DataSourceStream stream = new DataSourceStream(in);
            DummyMarkerListener mark = new DummyMarkerListener();
            client.put("/"+url.getPath()+"/"+name,stream,mark);
        } 
        // Johan: Added EOFException
        catch(java.io.EOFException err) {}
        catch(java.io.IOException err) {
            Disconnect();
            throw new ARCGridFTPJobException("IO error: "+err.getMessage());
        } catch(ServerException err) {
            Disconnect();
            throw new ARCGridFTPJobException("Communication with server failed: "+err.getMessage());
        } catch(ClientException err) {
            Disconnect();
            throw new ARCGridFTPJobException("Local problems: "+err.getMessage());
        }
    }

    public void Download(OutputStream out,String name) throws ARCGridFTPJobException {
        if(client == null) Connect();
        try{
            client.setPassiveMode(true);
            DataSinkStream stream = new DataSinkStream(out);
            DummyMarkerListener mark = new DummyMarkerListener();
            client.get("/"+url.getPath()+"/"+name,stream,mark);
        } 
        // Added by Johan, EOFException is considered harmless
        catch(java.io.EOFException err) { }
        catch(java.io.IOException err) {
            Disconnect();
            throw new ARCGridFTPJobException("IO error: "+err.getMessage());
        } catch(ServerException err) {
            Disconnect();
            throw new ARCGridFTPJobException("Communication with server failed: "+err.getMessage());
        } catch(ClientException err) {
            Disconnect();
            throw new ARCGridFTPJobException("Local problems: "+err.getMessage());

        }
    }

    public void List(String dir,Vector dirs,Vector files) throws ARCGridFTPJobException {
        if(client == null) Connect();
        Vector list = null;
        try {
              String fdir = new String("/"+url.getPath());
              if(!dir.equals("")) fdir+="/"+dir;
            client.setPassiveMode(true);
            list = client.mlsd(fdir);
            for(int i = 0;i<list.size();i++) {
                MlsxEntry entry = (MlsxEntry)(list.get(i));
                String type = entry.get("type");
                String name = entry.getFileName();
                if(name.equals(".")) continue;
                if(name.equals("..")) continue;
                if(!dir.equals("")) name=dir+"/"+name;
                if(type.equals("dir")) {
                  dirs.addElement(name);
                } else if(type.equals("file")) {
                  files.addElement(name);
                }
            }
        } catch(java.io.IOException err) {
            Disconnect();
            throw new ARCGridFTPJobException("IO error: "+err.getMessage());
        } catch(ServerException err) {
            Disconnect();
            if(err.getCode() != ServerException.SERVER_REFUSED) {
                throw new ARCGridFTPJobException
					("Communication with server failed: "+err.getMessage());
            } else {
            	logger.error("ERROR: "+err.toString());
            }
        } catch(org.globus.ftp.exception.ClientException err) {
            Disconnect();
            throw new ARCGridFTPJobException("Local problems: "+err.getMessage());
        }

    }

}

