/*
 * Created on Feb 23, 2006
 *
 * TODO Implement it, then write some documentation
 *
 */
package se.umu.jobsubmission.util;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.RandomAccessFile;
import java.io.StringWriter;
import java.util.Calendar;
import java.util.List;
import java.util.Vector;

import org.ggf.schemas.jsdl._2005._11.jsdl.JobDefinition_Type;
import org.ggf.schemas.jsdl._2005._11.jsdl.DataStaging_Type;
import org.globus.gsi.GlobusCredential;
import org.globus.wsrf.encoding.DeserializationException;
import org.globus.wsrf.encoding.ObjectDeserializer;
import org.globus.wsrf.encoding.ObjectSerializer;
import org.xml.sax.InputSource;

import se.umu.jobsubmission.Benchmark;
import se.umu.jobsubmission.JobObjectiveType;
import se.umu.jobsubmission.JobPreferencesType;
import se.umu.jobsubmission.JobRequirementType;
import se.umu.jobsubmission.SchedulingHintsType;
import se.umu.jobsubmission.client.ClientJobStaging;
import se.umu.jobsubmission.impl.datamanager.DataHelper;

/**
 * @author tordsson
 *
 * An utility class that contains some useful methods that helps clients
 * (and sometimes services) to perform simple job management tasks.
 *  
 */
public class ClientUtils {
	

    
    /**
     * Performs client-side file staging. When doing this, 
     * the JSS will not stage any input files. Client side file staging is done 
     * unless the user explicitly requests otherwise. 
     * @param jobID
     * @param jobDesc
     * @throws Exception
     */
    public static boolean stageJobInput(String jobID, 
    		JobDefinition_Type jobDesc, 
			String stagingTargetResolverClasses[], 
			String clientHost) throws Exception {
    	
    	List localInputFiles = null; 
    	
    	// Check if any job input files must be staged (at all) 
    	localInputFiles = DataHelper.getLocalInputFiles(jobDesc); 
    	if(localInputFiles == null) {
    		return false; 
    	}
    	
    	// Find out where to stage job input files. 
    	DataStaging_Type staging[] = jobDesc.getJobDescription().getDataStaging(); 
    	if(staging == null) { 
    		return true; // no input files to stage...     	
    	}
    	
    	// Save target directory for each inputfile
    	String filePath = null; 
    	
    	boolean done = false; 
    	for(int j = 0; j < stagingTargetResolverClasses.length && !done; j++) {
    		try {
    			ClientJobStaging stagingResolver = (ClientJobStaging) Class.forName(
    					stagingTargetResolverClasses[j]).newInstance(); 
    			filePath = stagingResolver.getJobDirectory(jobID, jobDesc); 
    			done = true; 
    		} catch (Exception e) {
    			System.err.println("Could not use class " + stagingTargetResolverClasses[j] + 
    					" to stage input: " + e.getMessage());													
    		}
    	} 
    	
    	// Do not know where to stage input files -> cancel job. 
    	if(filePath == null) {
    		System.err.println("Can not determine where to stage input files"); 
    		throw new Exception("Can not determine where to stage input files"); 
    	}
    	
    	GlobusCredential cred = GlobusCredential.getDefaultCredential(); 
    	// Perform actual staging (cancel job if staging fails). 
    	DataHelper.stageLocalInputFiles(filePath, localInputFiles, 
    			clientHost, cred); 
    	
    	return true; 
    }
    
    
    /**
     * Cancel job if file staging failes. This is handled by the same
     * plugin that handles resolving of staging target URL. 
     * @param jobId
     */
    public static void cancelJob(String jobId, 
    		String stagingTargetResolverClasses[]) {
    	boolean done = false; 
    	System.out.println("Cancelling job..."); 
    	for(int i = 0; i < stagingTargetResolverClasses.length && !done; i++) {
    		try {
    			ClientJobStaging stagingResolver = (ClientJobStaging) Class.forName(
    					stagingTargetResolverClasses[i]).newInstance(); 
    			stagingResolver.cancelJob(jobId); 
    			done = true; 
    			System.out.println("... job cancelled"); 
    		} catch (Exception e) {
    		}
    	}
    	if(!done) {
    		System.err.println("... can not cancel job"); 
    	}
    	
    }
    
    /**
     * Returns the list of job description translators as specified
     * in the client configuration file
     * @param clientConfig
     * @return
     */
    public static JobDescriptionTranslator[] getConverter(String clientConfig) {
    	
    	try {
    		BufferedReader in = new BufferedReader(new FileReader(clientConfig));
    		List classNames = new Vector(); 
    		while(in.ready()) {
    			String className = in.readLine().trim();
    			classNames.add(className); 
    		}
    		in.close();
    		JobDescriptionTranslator converters[] = 
    			new JobDescriptionTranslator[classNames.size()]; 
    		
    		for(int i = 0; i < classNames.size(); i++) {
    			converters[i] = (JobDescriptionTranslator) 
				Class.forName((String)classNames.get(i)).newInstance(); 
    		}
    	
    		return converters; 
    	} catch (Exception e) {
    		System.err.println("Can not create converter:"); 
    		return null; 
    	}
    }
    
    /**
     * Convert job description to JSDL, 
     * the client config file determines which converter to use. 
     */
    public static JobDefinition_Type convertJobDescription(String fileName, 
    		String converterClasses[]) throws FileNotFoundException, Exception {
    	  
    	// Try first for file existence
    	try {
    		RandomAccessFile f = new RandomAccessFile(fileName, "r");
    		f.close(); 
    	} catch (FileNotFoundException e) {
    		throw new FileNotFoundException ("Job description file " + fileName 
    				+ " not found"); 
    	}
    	
    	for(int i = 0; i < converterClasses.length; i++) {
    		JobDescriptionTranslator converter = null; 
    		try {
    			converter = (JobDescriptionTranslator) Class.forName(
    					converterClasses[i]).newInstance(); 
    			
    			JobDefinition_Type jobDef = (JobDefinition_Type) converter.
						translate(fileName); 
    			System.out.println("Translated job description " + fileName + 
    					" using " + converter.getClass().getName()); 
    			return jobDef; 
    		} catch (Exception e) {
    			System.out.println("Could not use " + converter.getClass().getName()
    					+ " to convert job description " + fileName); 
    		}
    	}
    	throw new Exception("No valid converter exists for " + fileName); 
    }
    
	
	 /**
    * Serializes job description and prints it to System.out.
    * This method is only used for debug purposes. 
    * @param job_desc
    */
   public static void printJobDescription(JobDefinition_Type job_desc) 
   	throws Exception {
   	
   	StringWriter sw = new StringWriter(); 
   	ObjectSerializer.serialize(sw, 
   			job_desc, 
				JobDefinition_Type.getTypeDesc().getXmlType()); 
   	System.out.println("Job description:");
   	System.out.println(sw.toString()); 
   	
   }
   

   
   /**
    * Reads and parses the job preferences file. If everything works
    * as expected, add the preferences file to the jobrequest. 
    * @param request
    * @param fileName
    */
   public static JobPreferencesType getJobPreferences(String fileName) {
   	
   	if(fileName != null) {
   		
   		try {
   			JobPreferencesType preferences = null; 
   			FileInputStream input = new FileInputStream(fileName); 
   			InputSource inputSource = new InputSource(input); 
   			preferences = (JobPreferencesType) ObjectDeserializer.deserialize( 
   					inputSource, JobPreferencesType.class);  
   			
   			ClientUtils.printPreferences(preferences);
   			
   			return preferences; 
   			
   		} catch (FileNotFoundException e) {
   			System.err.println("Job preferences file " + 
   					fileName + 
   			" not found, no job preferences will be used"); 
   		} catch(DeserializationException e) {
   			System.err.println("Deserialization error:\n" + e.getMessage()); 
   		} catch (Exception e) {
   			System.err.println("Could not read job preferences from file "
   					+ fileName); 
   		}
   		return null; 
   	} 
   	else {
   		System.err.println("No job preferences file specified"); 
   		return null; 
   	}
   }
   
   /**
    * Prints the job preferences, only used for debug purposes. 
    * @param prefs
    */
   public static void printPreferences(JobPreferencesType prefs) {
   	    	
   	if(prefs == null) {
   		System.out.println("No preferences specified"); 
   		return; 
   	}
   	
   	SchedulingHintsType hints = prefs.getSchedulingHints(); 
   	
   	if(hints != null) {
   		System.out.println("Scheduling hints:");
   		Benchmark benchmarks[] = hints.getBenchmark(); 
   		if(benchmarks != null) {
   			int size = benchmarks.length; 
   			for(int i = 0; i < size; i++) {
   				System.out.println("Benchmark: " 
   						+ benchmarks[i].getName() + " "
							+ benchmarks[i].getResult() + " "
							+ benchmarks[i].getTime()); 
   			}
   		}
   	
   	    if(hints.getFileTransfers() != null) {
   		System.out.println("Filetransfer times : in / out: " +
   			hints.getFileTransfers().getStageInTime() +
				" / " + hints.getFileTransfers().getStageOutTime());
   	    }
   	}
   	
   	JobRequirementType reqs = prefs.getJobRequirement();
   	if(reqs != null) {
   		System.out.println("Job requirements:");    	    	
   		
   		System.out.print("EarliestStart: "); 
   		Calendar c = reqs.getEarliestAbsoluteStart(); 
   		if(c != null)    		
   			System.out.println(c.getTime());
   		else
   			System.out.println(reqs.getEarliestRelativeStart());    	
   		
   		System.out.print("LatestStart: "); 
			c = reqs.getLatestAbsoluteStart(); 
			if(c != null)
				System.out.println(c.getTime()); 
			else
				System.out.println(reqs.getLatestRelativeStart()); 
   		
			Boolean useRes = reqs.getUseReservations(); 
			if(useRes == null || useRes.booleanValue()) {
								
				Boolean flex = reqs.getFlexible();
				if(flex != null && flex.booleanValue())
					System.out.println("Flexible reservation requested");
				else 
					System.out.println("Fixed reservation requested"); 
			}
			else {
				System.out.println("No reservation requested"); 
			}
			
   		JobObjectiveType objective = reqs.getJobObjective(); 
   		if(objective.equals(JobObjectiveType.earliestJobStart)) {
   			System.out.println("Earliest possible job start requested"); 
   		}
   		else if(objective.equals(JobObjectiveType.earliestJobCompletion))
   			System.out.println("Earliest possible job completion requested"); 
   		
   		String targets[] = reqs.getSubmissionTarget(); 
   		if(targets != null) {
   			System.out.println("Possible submission targets:"); 
   			for(int i = 0; i < targets.length; i++)
   				System.out.println(targets[i]); 
   		}
   		
   	}
   	    	    	    
   }

    

}
