/*
 * Created on Feb 23, 2006
 *
 * TODO Implement it, then write some documentation
 *
 */
package se.umu.jobsubmission.util;

import java.util.Calendar;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.Vector;

import org.globus.wsrf.security.SecurityManager;

import javax.xml.soap.SOAPBody;
import javax.xml.soap.SOAPElement;
import javax.xml.soap.SOAPEnvelope;
import javax.xml.soap.SOAPMessage;
import javax.xml.soap.SOAPPart;

import javax.security.auth.Subject;


import se.umu.jobsubmission.JobPreferencesType;
import se.umu.jobsubmission.JobRequest;
import se.umu.jobsubmission.JobRequirementType;
import se.umu.jobsubmission.NoAptResourcesFoundFaultType;
import se.umu.jobsubmission.NoResourcesFoundFaultType;
import se.umu.jobsubmission.impl.broker.Broker;
import se.umu.jobsubmission.impl.datamanager.DataManager;
import se.umu.jobsubmission.impl.datamanager.DefaultDataManager;
import se.umu.jobsubmission.impl.dispatcher.Dispatcher;
import se.umu.jobsubmission.impl.informationfinder.InformationFinder;

import org.apache.axis.Message;
import org.apache.axis.MessageContext;
import org.apache.axis.message.addressing.EndpointReferenceType;
import org.apache.log4j.Logger;
import org.ggf.schemas.jsdl._2005._11.jsdl.JobDefinition_Type;
import org.globus.delegation.DelegationUtil;
import org.globus.delegation.service.DelegationResource;
import org.globus.gsi.GlobusCredential;
import org.globus.gsi.gssapi.GlobusGSSCredentialImpl;
import org.globus.gsi.jaas.JaasSubject;
import org.globus.mds.glue.GLUECERPType;
import org.globus.wsrf.encoding.ObjectDeserializer;
import org.globus.wsrf.utils.FaultHelper;
import org.oasis.wsrf.faults.BaseFaultType;

import org.globus.wsrf.impl.security.authentication.Constants;


/**
 * @author tordsson
 *
 * This class contains a few convenience methods for usage
 * by both the coallocator and the submitter. 
 *
 */
public class SubmitterUtils {
	
	private static Logger logger = Logger.getLogger(SubmitterUtils.class); 
	
	 /**
     * Ensure that start time delay is expressed as an absolute time, 
     * convert it to absolute time if user specified a relative time.  
     * 
     * Do this for both earliest and latest start. 
     */
    public static JobPreferencesType fixPreferences(JobPreferencesType pref) {

        if(pref == null)
            return null;    		

        JobRequirementType jobRequirement = pref.getJobRequirement(); 

        if(jobRequirement == null)
            return pref;

        // EarliestStart
        Calendar absolute = jobRequirement.getEarliestAbsoluteStart(); 
        Integer relative = jobRequirement.getEarliestRelativeStart(); 
        if(absolute == null) {
            logger.debug("No absolute earliest start time requested"); 
            if(relative != null) {
                logger.debug("Relative earliest start time requested-> swap"); 
                Calendar now = Calendar.getInstance();
                now.add(Calendar.MINUTE, relative.intValue()); 
                jobRequirement.setEarliestRelativeStart(new Integer(0));
                jobRequirement.setEarliestAbsoluteStart(now); 
            } 
            else 
                logger.debug("No earliest job start time requested"); 
        }

        // LatestStart
        absolute = jobRequirement.getLatestAbsoluteStart(); 
        relative = jobRequirement.getLatestRelativeStart();  
        if(absolute == null) {
            if(relative != null) {
                logger.debug("Relative latest start requested -> swap"); 
                Calendar now = Calendar.getInstance(); 
                now.add(Calendar.MINUTE, relative.intValue()); 
                jobRequirement.setLatestRelativeStart(new Integer(0)); 
                jobRequirement.setLatestAbsoluteStart(now); 					
            }
            else 
                logger.debug("No latest job start time requested"); 
        }

        return pref;

    }
    

    /**
      Returns an array with instances of each valid Dispatcher. */
    public static Dispatcher[] getDispatchers(String dispatcherClasses[]) 
    	throws Exception {
    	List objects = new Vector(); 
    	for(int i = 0; i < dispatcherClasses.length; i++) {
    		try {
    			objects.add((Dispatcher) Class.forName(
    					dispatcherClasses[i]).newInstance()); 
    		} catch (Exception e) {
    			logger.warn("Error creating dispatcher of class: "
    					+ dispatcherClasses[i]); 
    		}
    	}
    	Dispatcher[] dispatchers = 
    		new Dispatcher[objects.size()]; 
    	for(int i = 0; i < objects.size(); i++) {
    		dispatchers[i] = (Dispatcher) objects.get(i); 
    	}
    	
    	return dispatchers; 
    }

    /**
      Return an array with instances of each valid DataManager. */
    public static DataManager[] getDataManagers(String dataManagerClasses[]) 
    	throws Exception {
    	List managers = new Vector(); 
    	for(int i = 0; i < dataManagerClasses.length; i++) {
    		try {
    			managers.add((DataManager) Class.forName(
    					dataManagerClasses[i]).newInstance()); 
    		} catch (Exception e) {
    			logger.warn("Error creating datamanager of class: "
    					+ dataManagerClasses[i]); 
    		}
    	}
    	DataManager dataManagers[] = 
    		new DataManager[managers.size() + 1]; 
    	for(int i = 0; i < managers.size(); i++) {
    		dataManagers[i] = (DataManager) managers.get(i); 
    	}
        dataManagers[dataManagers.length - 1] = new DefaultDataManager(); 
        return dataManagers; 
    }
    
    /**
     * Performs resource discovery (unless predefined submission targets 
     * are specified). Queries resources of interest for detailed information
     * @param informationFinder
     * @param jobDesc
     * @param indexServerURIs
     * @param jobPreferences
     * @param broker
     * @return
     * @throws BaseFaultType
     */
    public static List getResources(InformationFinder informationFinder, 
			String indexServerURIs[], 
			String resourceURIs[], 
			Broker broker) throws BaseFaultType {
 
    	// Find Grid resources
    	logger.debug("Finding resources..."); 				
    	List resources = null; 
    	List indexServers = new Vector(); 
    	
    	// Discovery, and query resources. 
    	if(resourceURIs == null) {
    		logger.info("Performing resource discovery"); 
    		if(indexServerURIs == null || indexServerURIs.length == 0)
    			indexServers = null; 
    		else {
    			int noIndexServers = indexServerURIs.length; 
    			for(int i = 0; i < noIndexServers; i++) 
    				indexServers.add(indexServerURIs[i]); 
    		}
    		
    		try {
    			resources = informationFinder.getResourceInfo(indexServers, 
    					true); 
    		} catch(Exception e) {
    			logger.error("No resource information retrieved from "); 
    			for(int i = 0; i < indexServerURIs.length; i++) {
    				logger.error(indexServerURIs[i]); 
    			}
    			NoResourcesFoundFaultType f = new NoResourcesFoundFaultType(); 
    			FaultHelper helper = new FaultHelper(f); 
    			String faultString = "No resource information retrieved from:\n"; 
    			for(int i = 0; i < indexServerURIs.length; i++) {
    				faultString += indexServerURIs[i]; 
    			}
    			helper.addDescription(faultString); 
    			throw helper.getFault(); 
    		}
    	}
    	// Resources predefined, just query them. 
    	else {
    		logger.info("Using predefined resources"); 
    		try {
    			for(int i = 0; i < resourceURIs.length; i++) {
    				indexServers.add(resourceURIs[i]); 
    			}
    			resources = informationFinder.getResourceInfo(indexServers, 
    					false); 
    		} catch (Exception e) {
    			logger.error("No resource information retrieved from "); 
    			String faultMessage = "No resource information retrieved from:\n"; 
    			for(int i = 0; i < resourceURIs.length; i++) {
    				logger.error(resourceURIs[i]); 
    				faultMessage += resourceURIs[i] + "\n"; 
    			}
    			// throw e; 
    			NoResourcesFoundFaultType f = new NoResourcesFoundFaultType(); 
    			FaultHelper helper = new FaultHelper(f); 
    			helper.addDescription(faultMessage); 
    			throw helper.getFault(); 
    		}
    	}
    	
    	logger.debug("Finding resources... done"); 
    	if(informationFinder != null)
    		informationFinder.shutdown(); 
    	
    	// TODO: Is this code ever reached? 
    	if(resources == null || resources.size() == 0) {        	
    		NoResourcesFoundFaultType f = new NoResourcesFoundFaultType(); 
    		FaultHelper helper = new FaultHelper(f); 
    		helper.addDescription("No resources found, can not submit job."); 
    		throw helper.getFault(); 
    	}
    	
    	return resources; 
    	
    }
    
    /**
     * Filters the found resources with respect to the constraints
     * expressed in the job description. 
     * @param resources
     * @param jobDesc
     * @param broker
     * @return
     * @throws Exception
     */
    public static List filterResources(List resources, 
    		JobDefinition_Type jobDesc, 
			Broker broker) throws Exception {
    	
    	// Filter out the resources that are not usable by the job. 
    	logger.info("Filtering the " + resources.size() + " found resources");
    	List approvedResources = broker.filter(resources, jobDesc);
    	logger.info("We have " + approvedResources.size() + 
    	" resources that are approved for the job ");
    	for(int j = 0; j < approvedResources.size(); j++) {
    		String name = ((GLUECERPType) approvedResources.get(j)).
			getCluster(0).getSubCluster(0).getHost(0).getName(); 
    		logger.debug(name); 
    	}
    	
    	if(approvedResources.size() == 0) {            
    		NoAptResourcesFoundFaultType f = new NoAptResourcesFoundFaultType(); 
    		FaultHelper helper = new FaultHelper(f); 
    		helper.addDescription("No suitable resource was found for the job"); 
    		throw helper.getFault(); 
    	}
    	
    	return approvedResources; 
    	
    }
    
    
	// Extract JSS JobRequest from MessageContext
	public static JobRequest getJobRequest(MessageContext context) {
		JobRequest jobRequest = null;
		try {	
			
		       Message message = context.getCurrentMessage();
               SOAPPart soapPart = message.getSOAPPart();
               SOAPEnvelope envelope = soapPart.getEnvelope();
               SOAPBody soapBody = envelope.getBody();
               Iterator bodyElements = soapBody.getChildElements();
               if (bodyElements.hasNext()) {
                   jobRequest = (JobRequest) ObjectDeserializer.toObject(
                       (SOAPElement) bodyElements.next(), JobRequest.class
                   );
               }
		} catch (Exception e) {
			logger.error("Failed to dig out soap message parameters: " + e);
		}
		return jobRequest;
	}
	
    /**
     * Gets the globus credential used by the service, 
     * if user specified GSISecureConversation with delegation, 
     * the delegated credential will be returned. 
     * 
     * If GSISecureMessage is used, and an epr to a delegated credential
     * is included in the job request, the delegated credential will be 
     * returned. 
     * @return
     * @throws Exception
     */
    public static GlobusCredential getDelegatedCredential() 
    	throws Exception {    	
    	
		MessageContext context = MessageContext.getCurrentContext(); 
		JobRequest jobRequest = getJobRequest(context); 
		EndpointReferenceType epr = jobRequest.getDelegatedCredentialEPR(); 

                // Credential delegated using DelegationService
		if(epr != null) {
			// Warning: This can only be called if a messageContext 
			// is set, i.e., from within the thread executing the service call
			logger.debug("Getting credential delegated using DelegationService"); 
			SecurityManager manager = SecurityManager.getManager(); 
			String DN = manager.getCaller(); 
			DelegationResource resource = DelegationUtil.getDelegationResource(epr); 
			GlobusCredential credential = resource.getCredential(); 
			return credential; 
		}
                // Credential delegated using secure-conversation with delegation. 
		else {
			logger.debug("Getting credential delegated using GSISecureConversation"); 
			Subject caller = (Subject)context.getProperty(Constants.PEER_SUBJECT);			
			Set s = caller.getPrivateCredentials(); 
			Iterator it = s.iterator(); 
    		GlobusCredential globusCreds = null; 
    		GlobusGSSCredentialImpl cred = (GlobusGSSCredentialImpl) it.next(); 
    		globusCreds = cred.getGlobusCredential(); 
    		return globusCreds; 
    		
			
    	} 
    }
    
    
    
    
    
}
